
#ifndef __FFTWUTL__
#define __FFTWUTL__
/** FFTW User Class
 * @Version $Rev: 200 $
 * @Author M.Sekido
 * @Date  $Date: 2016-08-28 22:40:41 +0900 (Sun, 28 Aug 2016) $
 * $Id: fftw_utl.hpp 200 2016-08-28 13:40:41Z vlbi $ 
 */

#include <iostream>
#include <vector>
// using namespace std;
#include "cmplx.hpp"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cmath>

extern "C" {
#include <fftw3.h>
}

inline std::vector <float> Blackman_Window(const int n){
  std::vector <float> arr(n);
  const double 	TWOPI =	2.0 * M_PI ;
  float dx = 1./(float)n;
  
  for(int i=0; i<n; i++){
    float x = (float)i*dx;
    arr[i] = 0.42 - 0.5*cosf(TWOPI*x) 
      + 0.08 * cos(2.0*TWOPI*x);
  }
  
  return arr;
}


class Fftwf{
public:
  enum Window_t {NA, BLACKMAN};
private:
  fftwf_complex *in, *out;
  fftwf_plan pf, pb;
  long n[3]; 
  int dim;
  Window_t win_t_; ///<< Window Type, Only one Dim at present
 public:
  /** Constructor */
  Fftwf(void):in(0),out(0),dim(0),win_t_(NA){n[0]=n[1]=n[2]=0;}
  /** Constructor 
   * \param ndat Number of data points for FFT
   * \param dimension Dimension for FFT						\
   * \param ndat2 2nd Number of point for 2D FFT.
   */
  Fftwf(long ndat, int dimension=1, int ndat2=0):in(0),out(0),dim(0),
						 win_t_(NA){
    Init(ndat,dimension,ndat2); };
  ~Fftwf();
  void Init(long ndat, int d=1, int ndat2=0);
  /** Execution of FFT */
  void Exec(int i){  
    
    if(win_t_ == BLACKMAN){
      vector <float> warr = Blackman_Window(n[0]);
      fftwf_complex *pc = in;
      switch(dim){
      case 1 :
	for(int i=0; i<n[0]; i++){ (*pc)[0] *= warr[i]; (*pc)[1] *= warr[i];  pc++;}
	break;
      case 2 :
	for(int j=0; j<n[1]; j++){
	  for(int i=0; i<n[0]; i++){ (*pc)[0] *= warr[i]; (*pc)[1] *= warr[i]; pc++;}
	}
	break;
      default:
	break;
      }
    }
    if(i == -1)
      fftwf_execute(pf); 
    else
      fftwf_execute(pb);
  
  }
  fftwf_complex* In(void){ return(in);}
  fftwf_complex* Out(void){ return(out);}
  long* Ndat(void){return(n);} 
  /** Set data in FComplex*  */
  void  Set_in(const FComplex *ca);
  /** Set data in vector <FComplex> */
  void  Set_in(const std::vector <FComplex> &ca);  
  /** Set data in float*  */
  void  Set_in(const float *fa);
  /** Set data in double*  */
  void  Set_in(const double *da);
  void  Set_out(FComplex *ca);  
  /** Shift output 1-D complex array by npt/2 **/
  void  Shift_out_1(FComplex *out);
  /** Shift output 2-D complex array by npt1/2 and npt2/2**/
  void  Shift_out_2(FComplex *out2);
  /** Shift input 1-D complex array by npt/2 **/
  void  Shift_in_1(const FComplex *in1, long n1=0);
  /** Shift input 2-D complex array by npt1/2 and npt2/2**/
  void  Shift_in_2(FComplex *in2);
  /** Shift input 1-D complex array by n1/2 **/
  void  Shift_in_1(const float *in1, long n1=0);
  void out_Normalize(float afact=1);
  void Chk_Wisdom(const char *);
  int Write_Wisdom(const char *);
  void Zero_in();
  void Zero_out();
  void Set_Window(const Window_t wt_){win_t_ = wt_;}
};

inline  void Fftwf::Zero_in(){
  uint32_t ndat= n[0];
  if( n[1] > 0) ndat *= n[1];  
  bzero(in,ndat*sizeof(fftwf_complex));
}

inline void Fftwf::Zero_out(){
  uint32_t ndat= n[0];
  if(n[1] > 0) ndat *= n[1];  
  bzero(out,ndat*sizeof(fftwf_complex));
}

inline Fftwf::~Fftwf(){
    if(n[0]>0){
      fftwf_destroy_plan(pf);
      fftwf_destroy_plan(pb);
      fftwf_free(in);
      fftwf_free(out);
    }
}

inline void Fftwf::Init(long ndat, int d, int ndat2){
  //  fftwf_complex *pc;
  n[0] = ndat;
  n[1] = ndat2;
  dim = d;
  switch(dim){
  case 1:
    if(in != 0) fftwf_free(in);
    if(out != 0) fftwf_free(out);
    in = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) * n[0]);
    out = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) * n[0]);
    if( in == 0 || out == 0){
      fprintf(stderr,"Error in Fftwf: Failed to get memory for in/out(%lu)\n",
	      n[0]*sizeof(fftwf_complex));
      exit(1);
    }
    pf = fftwf_plan_dft_1d(n[0], in, out, FFTW_FORWARD, FFTW_ESTIMATE);
    pb = fftwf_plan_dft_1d(n[0], in, out, FFTW_BACKWARD, FFTW_ESTIMATE);
    memset(in,0,sizeof(fftwf_complex)*n[0]);
    memset(out,0,sizeof(fftwf_complex)*n[0]);
/*     pc=in; */
/*     for(i=0; i<n[0]; i++){ (*pc)[0] =0.; (*pc)[1] =0.0; pc++;} */
/*     memcpy(out, in, sizeof(fftwf_complex)*n[0]); */
    break;
  case 2:{
    in = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) * n[0]*n[1]);
    out = (fftwf_complex *)fftwf_malloc(sizeof(fftwf_complex) * n[0]*n[1]);
    if( in == 0 || out == 0){
      fprintf(stderr,"Error in Fftwf: Failed to get memory for in/out(%lu)\n",
	      n[0]*n[1]*sizeof(fftwf_complex));
      exit(1);
    }
    pf = fftwf_plan_dft_2d(n[0], n[1], in, out, FFTW_FORWARD, FFTW_ESTIMATE);
    pb = fftwf_plan_dft_2d(n[0], n[1], in, out, FFTW_BACKWARD, FFTW_ESTIMATE);
    long nd = n[1]*n[0];
    memset(in,0,sizeof(fftwf_complex)*nd);
    memset(out,0,sizeof(fftwf_complex)*nd);

/*     pc= in; */
/*     for(i=0; i<nd; i++){ (*pc)[0] =0.; (*pc)[1] =0.0; pc++;} */
/*     memcpy(out, in, sizeof(fftwf_complex)*nd); */
  }
    break;
  default:
    break;
  }
}

inline void Fftwf::out_Normalize(float afact){
  
  long ndat;
  float fact;

  switch (dim){
  case 1:
    fact = afact/(double)n[0];
    ndat = n[0];
    break;
  case 2:
    fact = afact/(double)(n[0]*n[1]);
    ndat = n[0]*n[1];
    break;
  default :
    break;
  };
  
  fftwf_complex *pc = out;
  for(int i=0; i<ndat; i++){
    (*pc)[0] *= fact;
    (*pc)[1] *= fact;
    pc++;
  }
  
}

/****************************************/
inline void Fftwf::Shift_in_1(const FComplex* in2, long n1){

  if( n1 ==0) n1 = n[0];
  long n_half = n1/2;
  
  for(int i=0; i<n_half; i++){
    in[i][0] = in2[i+n_half].Real();
    in[i][1] = in2[i+n_half].Imag();
    in[i+n[0] - n_half][0] = in2[i].Real();
    in[i+n[0] - n_half][1] = in2[i].Imag();
  }
}

/****************************************/
inline void Fftwf::Shift_in_1(const float* in2, long n1){
  if( n1 ==0) n1 = n[0];
  long n_half = n1/2;
  
  for(int i=0; i<n_half; i++){
    in[i][0] = in2[i+n_half];
    in[i][1] = 0.;
    in[i+n[0] - n_half][0] = in2[i];
    in[i+n[0] - n_half][1] = 0.;
  }
}

/****************************************/
inline void Fftwf::Shift_out_1(FComplex* out2){
  long n_half = n[0]/2;
  for(int i=0; i<n_half; i++){
    out2[i].Set_val(out[i+n_half][0],out[i+n_half][1]);
    out2[i+n_half].Set_val(out[i][0],out[i][1]);
  }
}

inline void Fftwf::Shift_in_2(FComplex* in2){ // 2-dimensional
  //  long i,j, n_half;
  FComplex *pc(0), *pc2(0);
  FComplex* cbuf = new FComplex [n[0]*n[1]];
  
  long n_half = n[0]/2;
  for(int i=0; i<n[1]; i++){
    pc = cbuf + i * n[0];
    pc2 = in2 + i * n[0];
    for(int j=0; j<n_half; j++){
      pc[j] = pc2[j+n_half];
      pc[j+n_half] = pc2[j];
    }
  }    
  n_half = (long)(n[1]/2)*n[0];
  
  pc = cbuf;
  for(int j=0; j<n_half; j++){
    in[j][0] = pc[j+n_half].Real();
    in[j][1] = pc[j+n_half].Imag();
    in[j+n_half][0] = pc[j].Real();
    in[j+n_half][1] = pc[j].Imag();
  }
  
  delete [] cbuf;
}

inline void Fftwf::Shift_out_2(FComplex* out2){ // 2-dimensional
  long i,j, n_half;
  fftwf_complex *pf;
  FComplex *pc;
  FComplex* cbuf = new FComplex [n[0]*n[1]];
  
  
  n_half = n[0]/2;
  for(i=0; i<n[1]; i++){
    pc = cbuf + i*n[0];
    pf = out + i*n[0];
    for(j=0; j<n_half; j++){
      pc[j].Set_val(pf[j+n_half][0], pf[j+n_half][1]);
      pc[j+n_half].Set_val(pf[j][0], pf[j][1]);
    }
  }    
  n_half = (long)(n[1]/2)*n[0];
  
  pc = cbuf;
  for(j=0; j<n_half; j++){
    out2[j] = pc[j+n_half];
    out2[j+n_half] = pc[j];
  }
  
  delete [] cbuf;
}



inline void Fftwf::Set_in(const FComplex *ca){
  long int i, ndat;
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_in dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  for(i=0; i<ndat; i++){
    in[i][0] = ca[i].Real();
    in[i][1] = ca[i].Imag();
  }
}

inline void Fftwf::Set_in(const std::vector <FComplex> &ca){
  
  if( ca.size() != (size_t)n[0]){
    std::cerr << "Error: size (" << ca.size() 
	 << ") of complex array in FFTW.Set_in() differes from expected ("
	 << n[0] << endl;
    exit(1);
  }
  for(int i=0; i<n[0]; i++){
    in[i][0] = ca[i].Real();
    in[i][1] = ca[i].Imag();
  }
}

inline void Fftwf::Set_in(const float *fa)
{
  long int i, ndat;
  
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_in dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  
  for(i=0; i<ndat; i++){
    in[i][0] = fa[i];
    in[i][1] = 0.0;
  }
}

inline void Fftwf::Set_in(const double *da){
  long i, ndat;
  
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_out dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  for(i=0; i<ndat; i++){
    in[i][0] = (float)da[i];
    in[i][1] = 0.0;
  }
}

inline void Fftwf::Set_out(FComplex *ca){
  long int i, ndat ;
  switch (dim){
  case 1:
    ndat = n[0];
    break;
  case 2 :
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_out dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }
  for(i=0; i<ndat; i++)
    ca[i] = FComplex(out[i][0], out[i][1]);
}

inline int Fftwf::Write_Wisdom(const char *wisdom_file){
  FILE *fp;
  
  if( (fp= fopen(wisdom_file,"w")) == NULL){
    std::cerr <<"Error: Failed to write open: "
	 <<  wisdom_file << endl;
    return(-1);
  }
  fftwf_export_wisdom_to_file(fp);
  fclose(fp);
  return(0);
}

inline void Fftwf::Chk_Wisdom(const char *wisdom_file)
{
  FILE *fp;
  if( ( fp = fopen(wisdom_file, "r")) != NULL ){
    if( fftwf_import_wisdom_from_file(fp))
      std::cerr << "Error: Cannot read wisdom file." << endl ;
    fclose(fp);
  }else{
    std::cerr << "Warnning: Cannot open wisdome file:"
	 << wisdom_file << endl;
  }
}

/// Double Precision FFTW Class 
class Fftw{
  fftw_complex *in, *out;
  fftw_plan pf, pb;
  long n[3]; 
  int dim;
 public:
  Fftw(long ndat, int dimension=1, int ndat2=0);
  ~Fftw();
  void Exec(int i){  
    if(i == -1)
      fftw_execute(pf); 
    else
      fftw_execute(pb);
  }
  fftw_complex* In(void){ return(in);}
  fftw_complex* Out(void){ return(out);}
  long* Ndat(void){return(n);} 
  void  Set_in(const FComplex *ca);
  void  Set_in(const std::vector <FComplex> &ca);  
  void  Set_in(const float *fa);
  void  Set_in(const double *da);
  void  Set_out(FComplex *ca);  
  void  Shift_out_1(FComplex *out);
  void  Shift_out_2(FComplex *out2);
  void  Shift_in_1(const FComplex *in1, long n1=0);
  void  Shift_in_2(FComplex *in2);
  void  Shift_in_1(const float *in1, long n1=0);
  void out_Normalize(float afact=1);
  void Chk_Wisdom(const char *);
  int Write_Wisdom(const char *);
};

inline Fftw::~Fftw(){
    if(n[0]>0){
      fftw_destroy_plan(pf);
      fftw_destroy_plan(pb);
      fftw_free(in);
      fftw_free(out);
    }
}

inline Fftw::Fftw(long ndat, int d, int ndat2){
  n[0] = ndat;
  n[1] = ndat2;
  dim = d;
  switch(dim){
  case 1:
    in = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * n[0]);
    out = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * n[0]);
    if( in == 0 || out == 0){
      fprintf(stderr,"Error in Fftw: Failed to get memory for in/out(%lu)\n",
	      n[0]*sizeof(fftw_complex));
      exit(1);
    }
    pf = fftw_plan_dft_1d(n[0], in, out, FFTW_FORWARD, FFTW_ESTIMATE);
    pb = fftw_plan_dft_1d(n[0], in, out, FFTW_BACKWARD, FFTW_ESTIMATE);
    memset(in,0,sizeof(fftw_complex)*n[0]);
    memset(out,0,sizeof(fftw_complex)*n[0]);
/*     pc=in; */
/*     for(i=0; i<n[0]; i++){ (*pc)[0] =0.; (*pc)[1] =0.0; pc++;} */
/*     memcpy(out, in, sizeof(fftw_complex)*n[0]); */
    break;
  case 2:{
    in = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * n[0]*n[1]);
    out = (fftw_complex *)fftw_malloc(sizeof(fftw_complex) * n[0]*n[1]);
    if( in == 0 || out == 0){
      fprintf(stderr,"Error in Fftw: Failed to get memory for in/out(%lu)\n",
	      n[0]*n[1]*sizeof(fftw_complex));
      exit(1);
    }
    pf = fftw_plan_dft_2d(n[0], n[1], in, out, FFTW_FORWARD, FFTW_ESTIMATE);
    pb = fftw_plan_dft_2d(n[0], n[1], in, out, FFTW_BACKWARD, FFTW_ESTIMATE);
    long nd = n[1]*n[0];
    memset(in,0,sizeof(fftw_complex)*nd);
    memset(out,0,sizeof(fftw_complex)*nd);
  }
/*     pc= in; */
/*     for(i=0; i<nd; i++){ (*pc)[0] =0.; (*pc)[1] =0.0; pc++;} */
/*     memcpy(out, in, sizeof(fftw_complex)*nd); */
    break;
  default:
    break;
  }
}

inline void Fftw::out_Normalize(float afact){
  
  long i, ndat;
  float fact;
  fftw_complex *pc;

  switch (dim){
  case 1:
    fact = afact/(double)n[0];
    ndat = n[0];
    break;
  case 2:
    fact = afact/(double)(n[0]*n[1]);
    ndat = n[0]*n[1];
    break;
  default :
    break;
  };
  
  pc = out;
  for(i=0; i<ndat; i++){
    (*pc)[0] *= fact;
    (*pc)[1] *= fact;
    pc++;
  }
  
}

/****************************************/
inline void Fftw::Shift_in_1(const FComplex* in2, long n1){
  if( n1 ==0) n1 = n[0];
  long n_half = n1/2;
  
  for(int i=0; i<n_half; i++){
    in[i][0] = in2[i+n_half].Real();
    in[i][1] = in2[i+n_half].Imag();
    in[i+n[0] - n_half][0] = in2[i].Real();
    in[i+n[0] - n_half][1] = in2[i].Imag();
  }
}

/****************************************/
inline void Fftw::Shift_in_1(const float* in2, long n1){
  if( n1 ==0) n1 = n[0];
  long n_half = n1/2;
  
  for(int i=0; i<n_half; i++){
    in[i][0] = in2[i+n_half];
    in[i][1] = 0.;
    in[i+n[0] - n_half][0] = in2[i];
    in[i+n[0] - n_half][1] = 0.;
  }
}

/****************************************/
inline void Fftw::Shift_out_1(FComplex* out2){
  long n_half = n[0]/2;
  
  for(int i=0; i<n_half; i++){
    out2[i].Set_val(out[i+n_half][0],out[i+n_half][1]);
    out2[i+n_half].Set_val(out[i][0],out[i][1]);
  }
}

inline void Fftw::Shift_in_2(FComplex* in2){ // 2-dimensional
  FComplex *pc, *pc2;

  FComplex* cbuf = new FComplex [n[0]*n[1]];
  
  long n_half = n[0]/2;
  for(long i=0; i<n[1]; i++){
    pc = cbuf + i * n[0];
    pc2 = in2 + i * n[0];
    for(long j=0; j<n_half; j++){
      pc[j] = pc2[j+n_half];
      pc[j+n_half] = pc2[j];
    }
  }    
  n_half = (long)(n[1]/2)*n[0];
  
  pc = cbuf;
  for(long j=0; j<n_half; j++){
    in[j][0] = pc[j+n_half].Real();
    in[j][1] = pc[j+n_half].Imag();
    in[j+n_half][0] = pc[j].Real();
    in[j+n_half][1] = pc[j].Imag();
  }
  
  delete [] cbuf;
}

inline void Fftw::Shift_out_2(FComplex* out2){ // 2-dimensional
  fftw_complex *pf;
  FComplex *pc;
  FComplex* cbuf = new FComplex [n[0]*n[1]];
 
  long n_half = n[0]/2;
  for(long i=0; i<n[1]; i++){
    pc = cbuf + i*n[0];
    pf = out + i*n[0];
    for(long j=0; j<n_half; j++){
      pc[j].Set_val(pf[j+n_half][0], pf[j+n_half][1]);
      pc[j+n_half].Set_val(pf[j][0], pf[j][1]);
    }
  }    
  n_half = (long)(n[1]/2)*n[0];
  
  pc = cbuf;
  for(long j=0; j<n_half; j++){
    out2[j] = pc[j+n_half];
    out2[j+n_half] = pc[j];
  }
  
  delete [] cbuf;
}



inline void Fftw::Set_in(const FComplex *ca){
  long ndat;
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_in dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  for(int i=0; i<ndat; i++){
    in[i][0] = ca[i].Real();
    in[i][1] = ca[i].Imag();
  }
}

inline void Fftw::Set_in(const std::vector <FComplex> &ca){
  
  if(ca.size() != (size_t)n[0]){
    std::cerr << "Error: size (" << ca.size() 
	 << ") of complex array in FFTW.Set_in() differes from expected ("
	 << n[0] << endl;
    exit(1);
  }
  for(int i=0; i<n[0]; i++){
    in[i][0] = ca[i].Real();
    in[i][1] = ca[i].Imag();
  }
}

inline void Fftw::Set_in(const float *fa)
{
  long int i, ndat;
  
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_in dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  for(i=0; i<ndat; i++){
    in[i][0] = fa[i];
    in[i][1] = 0.0;
  }
}

inline void Fftw::Set_in(const double *da){
  long i, ndat;
  
  switch( dim){
  case 1: 
    ndat = n[0];
    break;
  case 2:
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_out dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }    
  for(i=0; i<ndat; i++){
    in[i][0] = (float)da[i];
    in[i][1] = 0.0;
  }
}

inline void Fftw::Set_out(FComplex *ca){
  long int i, ndat ;
  switch (dim){
  case 1:
    ndat = n[0];
    break;
  case 2 :
    ndat = n[0]*n[1];
    break;
  default :
    fprintf(stderr,"Error: FFTW Set_out dimension error:(%d)\n", dim);
    exit(EXIT_FAILURE);
  }
  for(i=0; i<ndat; i++)
    ca[i] = FComplex(out[i][0], out[i][1]);
}

inline int Fftw::Write_Wisdom(const char *wisdom_file){
  FILE *fp;
  
  if( (fp= fopen(wisdom_file,"w")) == NULL){
    std::cerr <<"Error: Failed to write open: "
	 <<  wisdom_file << endl;
    return(-1);
  }
  fftw_export_wisdom_to_file(fp);
  fclose(fp);
  return(0);
}

inline void Fftw::Chk_Wisdom(const char *wisdom_file)
{
  FILE *fp;
  if( ( fp = fopen(wisdom_file, "r")) != NULL ){
    if( fftw_import_wisdom_from_file(fp))
      std::cerr << "Error: Cannot read wisdom file." << endl ;
    fclose(fp);
  }else{
    std::cerr << "Warnning: Cannot open wisdome file:"
	      << wisdom_file << endl;
  }
}

#endif
